/****************************************************************************/
/*	SED - RAM - Microcontroller interface by Arthur Computer Systems, 2005	*/
/*==========================================================================*/
/*	Target chip: XC95144XL-10-TQ100											*/
/*--------------------------------------------------------------------------*/
/*	A chip-re ra kell kotni 32 vagy 64 k 15..20 ns SRAM-ot, a SED-et,		*/
/*	50MHz clock-ot.															*/
/*	A SED SEL1 labat 1-be kell tenni, az A0 labara a CPUA[0]-t kell kotni,	*/
/*  WR labara a CPURnW jelet kell kotni, CS labat le kell foldelni, a RD 	*/
/*	jelere pedig a SEDE jelet kell kotni. Az adatbuszara a SEDD[7:0] jelek	*/
/*	mennek. A SED 10 MHz-ces orajelet ennek a chip-nek a CLKOUT labarol		*/
/*	lehet elvezetni.														*/
/*																			*/
/*	Ezen felul marad egy busz, amin a kovetkezo jelek vannak:				*/
/*																			*/
/*	CPUA[2:0] - 3 bit cimbusz 												*/
/*	CPUD[7:0] - 8 bit, ketiranyu adatbusz									*/
/*	nCPUCE    - aktiv low chip enable jel									*/
/*	CPURnW    - iras (low) / olvasas (high)									*/
/*																			*/
/*	A CPU ciklus ugy nez ki, hogy adni kell a cimet, az R/W-t, iras eseten	*/
/*	az adatot, es az nCPUCE jelet le kell huzni nullaba. Akar egyszerre is	*/
/*	beallhat minden jel az nCPUCE-vel egyutt, de semmikeppen nem kesobb!	*/
/*	A ciklus alatt nem valtozhat semelyik jel. A ciklus vegen eloszor az	*/
/*	nCPUCE-t kell egybe tenni, es a tobbi jelet legalabb 50 ns-ig hagyni.	*/
/*																			*/
/*	Ha a cimbuszon 0 vagy 1 van, akkor az adott iras vagy olvasas a SED		*/
/*	fele lesz tovabbitva, azaz ilyenkor a SED regisztereit lehet elerni.	*/
/*	Ebben az esetben az nCPUCE legalabb 150ns-ig aktiv kell legyen			*/
/*																			*/
/*	A 6 es 7 cimekre torteno irassal lehet a rwaddress regizter felso es	*/
/*	also byte-jat lehet beallitani.											*/
/*																			*/
/*	A 2 es 3 cimeken a RAM-ot lehet elerni, megpedig az rwaddress regiszter	*/
/*	szerinti cimen. Az cim=2 elereskor a az rwaddress regiszter eggyel		*/
/*	novekszik a muvelet utan. Ha a 3-mas cimen erjuk el a RAM-ot, akkor nem.*/
/*  																		*/
/*  RAM olvasaskor az nCPUCE jelet kovetoen legkesobb 140ns ido mulva 		*/
/*	jelenik meg a chip adatbuszan a RAM-bol felolvasott adat. Ebbol adodoan	*/
/*	egy RAM olvasasi ciklus minimum ennyi ideig kell tartson.				*/
/*	Irasnal az nCPUCE legalabb 80ns-ig aktiv kell legyen					*/
/*																			*/
/* Idozitesek osszefoglalva:												*/
/*	Cim, R/W, data => nCPUCE 0    min.  0ns	(setup)							*/
/*	nCPUCE 1 => cim, r/w, data    min. 50ns	(hold)							*/
/*	nCPUCE inaktiv allapot        min. 50ns									*/
/*  nCPUCE aktiv 															*/
/*    SED RD / WR                 min. 150ns								*/
/*    ramrwaddess WR              min.  40ns								*/
/*    RAM RD                      min. 140ns								*/
/*    RAM WR                      min.  80ns								*/
/*--------------------------------------------------------------------------*/
/* History:																	*/
/*	2005.08.26: First draft													*/
/*	2005.09.23: frame szamlalo, c000-efff maszkolo							*/
/*	2005.09.26: 'step' regiszter es tartozekai								*/
/****************************************************************************/

module	chip(CLK, CLKOUT,
			 nCPUCE,  CPURnW,  CPUA, CPUD,
			 nSEDVCE, nSEDVWR, nSEDVRD, SEDVA, SEDVD, SEDE, SEDD,
			 nRAMCE0, nRAMCE1, nRAMWR, nRAMOE, RAMA, RAMD, frame, test1, test2
);

input			CLK;		// 50 MHz orajel be
output			CLKOUT;		// 10 MHz orajel ki

input			nCPUCE;		// Chip enable, aktiv low
input			CPURnW;		// Iras/olvasa => SED R/W
input	[2:0]	CPUA;		// CPU cimbusz also 3 bit => SED A0
inout	[7:0]	CPUD;

input			nSEDVCE;	// SED RAM jelek
input			nSEDVWR;
input			nSEDVRD;
input	[15:0]	SEDVA;
inout	[7:0]	SEDVD;
output			SEDE;		// SED E
inout	[7:0]	SEDD;		// SED data bus

output			nRAMCE0;	// 27 i/o
output			nRAMCE1;
output			nRAMWR;
output			nRAMOE;
output	[14:0]	RAMA;
inout	[7:0]	RAMD;

input			frame;

output			test1;
output			test2;

reg				nRAMCE0 = 1;
reg				nRAMCE1 = 1;
reg				nRAMWR = 1;
reg				nRAMOE = 1;
reg		[14:0]	RAMA = 0;

reg		[1:0]	state = 0;

reg				ssce = 1;
reg		[1:0]	sce = ~0;
wire			sedfall;

reg		[1:0]	scce = ~0;
reg				cpufall1 = 0;
reg				cpuf1d = 0;
reg				cpufall3 = 0;

reg		[7:0]	ramwdat = 0;

reg				stosed = 0;
reg		[7:0]	sedrram = 0;

reg				stocpu = 0;
reg		[7:0]	cpurram;

reg		[2:0]	step = 0;
reg		[15:0]	cpuaddr = 0;
reg				cpuinc = 0;

reg		[2:0]	clkdiv = 0;
reg				clk1 = 0;
reg				clk2 = 0;


reg		[2:0]	cnt = 0;
reg		[2:0]	pres = 0;
wire			xarea;


	assign test1 = frame;
	assign test2 = nSEDVRD;

	always @ (negedge frame)
		if(cnt == pres)
			cnt <= 0;
		else
			cnt <= cnt + 1;


	//------------------------------------------------------------------
	//	Eloallitja a 10MHz-cet az 50-bol
	//------------------------------------------------------------------
	
	always @ (posedge CLK) begin
		if(clkdiv == 4) begin
			clkdiv <= 0;
			clk1   <= 1;
		end
		else
			clkdiv <= clkdiv + 1;
			
		if(clkdiv == 1)
			clk1 <= 0;
	end
	
	always @ (negedge CLK)
		clk2 <= clk1;
	
	assign CLKOUT = (clk1 | clk2) ? 1'bz : 1'b0;

	//------------------------------------------------------------------
	//
	// RAM fele meno adat meghajtasa (TS)
	//
	assign RAMD = (!nRAMWR) ? ramwdat : 8'hzz;
	
	
	// SED RAM adatbuszanak meghajtasa (TS)
	//
	assign SEDVD = (!nSEDVCE && !nSEDVRD) ? sedrram : 8'hzz;

	// Adatbusz mehajtasa a CPU fele (TS)
	//
	assign CPUD = (!nCPUCE && CPURnW && ((CPUA == 2) || (CPUA == 3))) ? cpurram :
				  (!nCPUCE && CPURnW && ((CPUA == 0) || (CPUA == 1))) ? SEDD :
																		8'hzz;

	// SED E jelenek eloallitasa
	//
	assign SEDE = (!nCPUCE && ((CPUA == 0) || (CPUA == 1)));
	
	// Adatbusz meghajtasa a SED fele (TS)
	//
	assign SEDD = (!nCPUCE && !CPURnW && ((CPUA == 0) || (CPUA == 1))) ? CPUD :
																		 8'hzz;
		
	assign xarea = (SEDVA >= 16'hc000) && (SEDVA <= 16'hEFFF);

	//-----------------------------------------------------------------------
	// SED nVCE jel mintavetelezese orajel lefuto elnel, es a masfel orajelig
	// tarto sedfall jel eloallitasa
	//-----------------------------------------------------------------------

	always @ (negedge CLK)
		ssce <= nSEDVCE;

	always @ (posedge CLK)
		sce <= (sce << 1) | ssce;
		
	assign sedfall = (((ssce == 1'b0) && (sce[0] == 1'b1)) ||
					  ((sce[0] == 1'b0) && (sce[1] == 1'b1)));

	//------------------------------------------------------------------
	// CPU nCE jel mintavetelezese orajel felfuto elenel egy ketbites shift-be
	// majd lefutonal eloall az 1 orajel hosszu cpufall1 es a 
	// harom orajel hosszu cpufall3, ami csak bizonyos esetekben aktiv!!!
	//------------------------------------------------------------------
	
	always @ (posedge CLK)
		scce <= (scce << 1) | nCPUCE;
		
	always @ (negedge CLK) begin
		cpufall1 <=   !scce[0] && scce[1];
		cpuf1d   <=  cpufall1;
		cpufall3 <= ((!scce[0] && scce[1]) || cpufall1 || cpuf1d) && 
		                           !stocpu && ((CPUA==2) || (CPUA==3));
	end

	
	//------------------------------------------------------------------
	//	Main state machine
	//------------------------------------------------------------------

	always @ (posedge CLK) begin

		//---------------------------------------------------------------
		// A fizikai RAM muveletek mindig ket orajelig tartanak
		// Az elsonel kitesz minden jelet, (CE, WR, OE, ADDR, DATA ha kell)
		// A masodik orajel elejen elveszi a WR-t es a DATA-t, ha iras volt
		// Es a harmadik orajelnel visszaveszi a jeleket, ha nincs tovabbi
		// muvelet, egyebkent kiteszi az uj CE, ADDR, ... jeleket.
		// Ha olvasta a RAM-ot, akkor a harmadik orajel elejen vesz mintat
		// a ram adatbol.
		//
		// Az abran van egy iras, rogton utana egy olvasas, majd egy orajel
		// szunet utan meg egy olvasas.
		//     
		//		 __    __    __    __    __    __    __    __    __
		// CLK	   \__/  \__/  \__/  \__/  \__/  \__/  \__/  \__/  \_
		//    		__                         _____             ____
		// nRAMCE  	  \_______________________/     \___________/
		//			__ ___________ _________________ ________________
		// RAMADDR	__X___________X_________________X________________
		//			__       ________________________________________
		// nRAMWR	  \_____/                           
		//			______________             ______            ____
		// nRAMOE	              \___________/      \__________/
		//             _____       ___________        __________
		// RAMDATA	--Xchip_X-----X____RAM____X------X___RAM____X-----
		//         ___ _____ _____ _____ _____ _____ _____ _____ _____   
		// state   0__X__1__X__2__X__1__X__2__X__0__X__1__X__2__X__0__
		//
		state <= state << 1;
		

		if(cpufall1 && (CPUA == 4)) begin
			step <= CPUD;
		end

		if(cpufall1 && (CPUA == 5)) begin
			pres <= CPUD;
		end
		
		if(cpufall1 && (CPUA == 6)) begin
			cpuaddr[15:8] <= CPUD;
		end
		
		if(cpufall1 && (CPUA == 7)) begin
			cpuaddr[7:0]  <= CPUD;
		end

		/*
		 *	Read / write cim allitasa, ha kell
		 */
		cpuinc <= 0;

		if(cpuinc) begin
			cpuaddr <= cpuaddr + ((step == 3'h0) ? 1 :	
						          (step == 3'h1) ? 16 :  // 128 pixel
								  (step == 3'h2) ? 20 :	 // 160 pixel
								  (step == 3'h3) ? 32 :  // 256 pixel
								  (step == 3'h4) ? 40 :  // 320 pixel
								  (step == 3'h5) ? 64 : 80); // 512, 640 pixel
		end
		

		if(state == 1) begin
			nRAMWR <= 1;
		end
		
		if(state == 2) begin
			nRAMCE0 <= 1;
			nRAMCE1 <= 1;
			nRAMWR  <= 1;
			nRAMOE  <= 1;
			
			if(stosed)
				sedrram <= (!nRAMOE) ? RAMD : 8'h00;
			
			if(stocpu)
				cpurram <= RAMD;
			
			stosed <= 0;
			stocpu <= 0;
		end
		
		if((state == 0) || (state == 2)) begin
			//
			//	RAM ciklus esetleges kezdete
			//
			if(sedfall) begin
				
				// SED akarja a RAM-ot, megkapja
				
				RAMA    <= SEDVA[14:0];
				nRAMCE0 <=  SEDVA[15];
				nRAMCE1 <= !SEDVA[15];
				nRAMWR  <= nSEDVWR;
				nRAMOE  <= 0; //!(!nSEDVWR && (!xarea || (cnt == 0)));
				ramwdat <= SEDVD;
				stosed  <= 1;
				state   <= 1;
			end
			else if(cpufall3) begin
				//
				// SED-nek nem kell, CPU-nak igen
				//
				RAMA    <= cpuaddr[14:0];
				nRAMCE0 <=  cpuaddr[15];
				nRAMCE1 <= !cpuaddr[15];
				nRAMWR  <=  CPURnW;
				nRAMOE  <= !CPURnW;
				ramwdat <= CPUD;
				cpuinc  <= (CPUA == 2);
				stocpu  <= 1;
				state   <= 1;
			end
		end
	end
endmodule

